/******************************************************************************
         Image Acquisition & Display Program - OEM Control Functions
                       Runs on any VGA or MCGA display.
******************************************************************************/
#include <bios.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <at200.h>
#include <at200oem.h>
#include <camera.h>

#define DEFAULT_CCD       4           /* Thomson 7895B CCD */
#define DEFAULT_ANALOG    A200KHZ     /* 200KHz 14 bit analog electronics */
#define DEFAULT_DATA_BITS 14          /* 200KHz 14 bit analog electronics */
#define DEFAULT_EXPOSURE  200         /* Exposure time in milliseconds */
#define DEFAULT_SHUTTER   31          /* Shutter open/close time in milliseconds */
#define DEFAULT_ADDRESS   2           /* Maps to AT200 at address 0x220 */

int ccd_type  = DEFAULT_CCD,
    ser_byte  = DEFAULT_ANALOG,
    data_bits = DEFAULT_DATA_BITS,
    shutter   = DEFAULT_SHUTTER,
    at200_id  = DEFAULT_ADDRESS;
long exposure = DEFAULT_EXPOSURE;

char far *screen_row[200] = { 0, };   /* Pointers to graphics display rows */
/* screen_row[] is initialized in main() to point to the start of each row on
   the VGA display.  It is then used by the show_image() function to update
   the VGA memory. */

unsigned huge buffer[65536] = { 0, }, /* Camera Image Buffer */
         far *ptr[200] = { 0, },      /* Image Display Pointers */
         far *ptable[32] = { 0,};     /* Image Collection Pointers */
/* buffer[] is an array of 64 K pixels (128 K bytes) that is used to collect
   data from the camera.  The ptable[] array is used by the AT200 control
   library functions to collect the pixels.  It is initialized in main() to
   point to each 2 K pixel (4 K byte) block in the buffer[] array.  The ptr[]
   array is used by show_image() to display 320 pixel rows on the VGA.  It is
   also initialized in main(). */

void mcga_display_mode(int mode)
{ /* Select text mode or 320 x 200 - 256 color graphics mode */
union REGS regs;

  regs.x.ax = (mode) ? 19 : 3;  /* AH=0, AL=Mode */
  int86(16, &regs, &regs);      /* Call video bios */
}

void mcga_color_set(int color,int red,int green,int blue)
{ /* Set the RGB values for a color in 256 color graphics modes */
  outp(0x3C8,color);  /* DAC color control register selects the color */
  outp(0x3C9,red);    /* Set Red DAC (0..63) */
  outp(0x3C9,green);  /* Set Green DAC (0..63) */
  outp(0x3C9,blue);   /* Set Blue DAC (0..63) */
}

void far *normalize_pointer(void far *data)
/* Make pointers have the lowest offset possible (and the largest segment).
   This function is needed to insure that each pointer in the ptable[] array
   can be used for 2 K pixels (4 K bytes).  Taking the address of an element
   of a huge array only guarantees that the segment:offset pair will be valid
   for that element.  The calculation performed here guarantees that the
   segment:offset pair will be valid for at least the 4 K addresses we need.
   See the Microsoft Macro Assembler 5.1 manual, section 13.2 for more
   information on REAL MODE ADDRESSING. */
{
unsigned seg,offs;

  seg  = (unsigned) ((long) data >> 16L);
  offs = (unsigned) ((long) data & 0xFFFF);
  seg += offs >> 4;
  offs &= 15;
  return( (void far *) (((long) seg << 16L) | (long) offs) );
}

void characterize_ccd(int frame)      /* Set up the CCD characterization */
{
int x;
cam_sys config;

  config.ccd = ccd[ccd_type];
  config.at200_id = at200_id;
  config.frame = frame;
  config.open_delay = shutter;
  config.close_delay = shutter;
  config.preamp_delay = 100;
  config.speed = ser_byte;
  config.data_bits = data_bits;
  x = setup(&config);                 /* Set up the CCD characterization */
  if( x ) {
    printf("setup ERROR %d: %s\n",x,error_msgs[x]);
    exit( 1 );
  }
}

void select_readout(int frame,int max_bin)
{ /* Set the exposure parameters & subarray */
int x,rows,p_bin,s_bin;
cam_exp readout;

  readout.ram = ptable;
  readout.exposure = exposure;
  readout.gain = 0;
  if( frame )
    rows = ccd[ccd_type].rows / 2;
  else
    rows = ccd[ccd_type].rows;
  if( max_bin ) {
    p_bin = rows / 200;
    s_bin = ccd[ccd_type].cols / 320;
  }
  else
    p_bin = s_bin = 1;
  readout.corg = (ccd[ccd_type].cols - 320 * s_bin) / 2;
  readout.cols = 320;
  readout.cbin = s_bin;
  readout.rorg = (rows - 200 * p_bin) / 2;
  readout.rows = 200;
  readout.rbin = p_bin;
  readout.clr_cnt = 0;
  readout.row_shift = ccd[ccd_type].rows / 2;
  x = set_exp(&readout);                          /* Set up the readout */
  if( x ) {
    printf("set_exp ERROR %d: %s\n",x,error_msgs[x]);
    exit( 1 );
  }
}

void show_image(void)   /* Display a 320 x 200 image on the VGA */
{
unsigned char buff[320];
unsigned max,min;
register int x,y;

  for( min=65535,max=y=0 ; y<200 ; y++ )
    for( x=319 ; x>=0 ; x-- ) {
      if( max < ptr[y][x] )
        max = ptr[y][x];
      if( min > ptr[y][x] )
        min = ptr[y][x];
    }
  max = (unsigned) (((long) max - (long) min + 62L) / 63L);
  if( !max )
    max = 1;    /* Don't divide by zero */
  for( y=0 ; y<200 ; y++ ) {
    for( x=0 ; x<320 ; x++ )
      buff[x] = (unsigned char) ((ptr[y][x] - min) / max);
    memcpy(screen_row[y],buff,320);
  }
}

void do_exposures(int frame,int max_bin)
{ /* Do a sequence of readouts in quick succession with display */
int x;

  characterize_ccd(frame);      /* Characterize the electronics */
  select_readout(frame,max_bin);/* Set the exposure parameters */
  if( frame )
    oshut();                    /* Open the shutter */
  x = expose_light();           /* Script to DSP, read out CCD first exposure */
  if( x == -1 )                 /* Long exposures cause early return */
    cam_wait();                 /* Collect the pixels */
  else if( x ) {
    printf("expose_light ERROR %d: %s\n",x,error_msgs[x]);
    if( frame )
      cshut();                  /* Close the shutter */
    exit( 1 );
  }
  mcga_display_mode( 1 );       /* Start Graphics Mode */
  for( x=0 ; x<64 ; x++ )       /* Set up palette (64 shades of gray) */
    mcga_color_set(x,x,x,x);
  while( !kbhit() ) {           /* Loop until a key is pressed */
    show_image();               /* Display the image on the VGA */
    if( !kbhit() ) {            /* If no key was pressed... */
      dsp_command( EXECUTE );   /* Re-use old script (from expose_light()) */
      collect_image(ptable);    /* Use the fast routine to collect Pixels */
    }
  }
  getch();                      /* Read the key that was presed */
  if( frame )
    cshut();                    /* Close the shutter */
  mcga_display_mode( 0 );       /* Start Text Mode */
}

void show_help(void)
{ /* Display the usage message for this program */
register x;
char buff[81],number[8];

  printf("USAGE: VIEW [-a #] [-b] [-c #] [-e #] [-f] [-o #] [-s #]\n"
         "  Where:\n"
         "    -a #  # is the base address descriptor (2 by default)\n"
         "    -b    indicates bin to see full field of view\n"
         "    -c #  # selects the CCD type from the following list:\n");
  strcpy(buff,"           ");
  for( x=0 ; x<CCD_COUNT ; x++ ) {
    sprintf(number," %u=",x);
    if( strlen(buff) + strlen(number) + strlen(ccd[x].id) < 80 ) {
      strcat(buff,number);
      strcat(buff,ccd[x].id);
    }
    else {
      printf("%s\n",buff);
      strcpy(buff,"           ");
      strcat(buff,number);
      strcat(buff,ccd[x].id);
    }
  }
  printf("%s\n",buff);
  printf("    -e #  # is the exposure time in milliseconds (>=0)\n"
         "    -f    indicates a frame transfer CCD\n"
         "    -o #  # is the open/close time for the shutter in milliseconds (>=0)\n"
         "    -r #  # is the resolution of the A/D converter in bits\n"
         "    -s #  # selects the A/D Converter speed where:\n"
         "            0=40 KHz, 1=50 KHz, 2=200 KHz, 3=500 KHz\n");
  exit( 0 );
}

void main(int argc,char *argv[])
{
register unsigned x;
int frame = 0,max_bin = 0;

  for( x=1 ; x < (unsigned) argc ; x++ ) {  /* Parse the command line */
    if( argv[x][0] == '-' )
      switch( argv[x][1] ) {
        case 'a': case 'A':
          x++;
          at200_id = atoi(argv[x]);
          break;
        case 'b': case 'B':
          max_bin = 1;
          break;
        case 'c': case 'C':
          x++;
          ccd_type = atoi(argv[x]);
          if( ccd_type >= CCD_COUNT || ccd_type < 0 )
            show_help();
          break;
        case 'e': case 'E':
          x++;
          exposure = atoi(argv[x]);
          if( exposure < 0 )
            show_help();
          break;
        case 'f': case 'F':
          frame = 1;
          break;
        case 'o': case 'O':
          x++;
          shutter = atoi(argv[x]);
          if( shutter < 0 )
            show_help();
          break;
        case 'r': case 'R':
          x++;
          data_bits = atoi(argv[x]);
          break;
        case 's': case 'S':
          x++;
          ser_byte = atoi(argv[x]);
          if( ser_byte < A40KHZ || ser_byte > A500KHZ )
            show_help();
          break;
        default:
          show_help();
          break;
      }
    else
      show_help();
  }
  /* Display the configuration */
  printf("Configuration:\n  %s CCD.\n  ",ccd[ccd_type].id);
  switch( ser_byte ) {
    case A40KHZ:  printf("40");  data_bits = 16; break;
    case A50KHZ:  printf("50");  data_bits = 14; break;
    case A200KHZ: printf("200"); break;
    case A500KHZ: printf("500"); data_bits = 12; break;
  }
  printf(" KHz Analog Card.\n");
  /* Set up the pointers used above */
  for( x=0 ; x<32 ; x++ )     /* Load image acquisition array */
    ptable[x] = normalize_pointer(&buffer[x * 2048]);
  for( x=0 ; x<200 ; x++ ) {  /* Load image and graphics display arrays */
    ptr[x] = normalize_pointer(&buffer[x * 320]);
    screen_row[x] = &((char far *)0xA0000000)[x * 320];
  }
  do_exposures(frame,max_bin);  /* Perform readout and display */
  exit( 0 );
}
